import os
from collections import OrderedDict
from contextlib import contextmanager

from medikit.config import read_configuration, load_features
from medikit.events import LoggingDispatcher, ProjectEvent
from medikit.feature import ProjectInitializer
from medikit.pipeline import ConfiguredPipeline
from medikit.settings import DEFAULT_FILES, DEFAULT_FEATURES

from mondrian import term

START = 'start'
CONTINUE = 'continue'
ABORT = 'abort'


def _read_configuration(dispatcher, config_filename):
    """
    Prepare the python context and delegate to the real configuration reader (see config.py)

    :param EventDispatcher dispatcher:
    :return tuple: (variables, features, files, config)
    """
    if not os.path.exists(config_filename):
        raise IOError('Could not find project description file (looked in {})'.format(config_filename))

    variables = OrderedDict()

    files = {filename: '' for filename in DEFAULT_FILES}
    features = set(DEFAULT_FEATURES)

    return read_configuration(dispatcher, config_filename, variables, features, files)


@contextmanager
def _change_working_directory(path):
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def handle_init(config_filename, **options):
    if os.path.exists(config_filename):
        raise IOError(
            'No config should be present in current directory to initialize (found {})'.format(config_filename)
        )

    config_dirname = os.path.dirname(config_filename)
    if not os.path.exists(config_dirname):
        os.makedirs(config_dirname)

    # Fast and dirty implementation
    # TODO
    # - input validation
    # - getting input from env/git conf (author...),
    # - dispatching something in selected features so maybe they can suggest deps
    # - deps selection
    # - ...
    with _change_working_directory(config_dirname):
        dispatcher = LoggingDispatcher()
        initializer = ProjectInitializer(dispatcher, options)
        initializer.execute()
        return handle_update(config_filename)


def handle_update(config_filename, **kwargs):
    import logging
    logger = logging.getLogger()

    dispatcher = LoggingDispatcher()

    variables, features, files, config = _read_configuration(dispatcher, config_filename)

    feature_instances = {}
    logger.info(
        'Updating {} with {} features'.format(
            term.bold(config['python'].get('name')),
            ', '.join(term.bold(term.green(feature_name)) for feature_name in sorted(features))
        )
    )

    all_features = load_features()

    sorted_features = sorted(features)  # sort to have a predictable display order
    for feature_name in sorted_features:
        logger.debug('Initializing feature {}...'.format(term.bold(term.green(feature_name))))
        try:
            feature = all_features[feature_name]
        except KeyError as exc:
            logger.exception('Feature "{}" not found.'.format(feature_name))

        if feature:
            feature_instances[feature_name] = feature(dispatcher)

            for req in feature_instances[feature_name].requires:
                if not req in sorted_features:
                    raise RuntimeError('Unmet dependency: {} requires {}.'.format(feature_name, req))

            for con in feature_instances[feature_name].conflicts:
                if con in sorted_features:
                    raise RuntimeError('Conflicting dependency: {} conflicts with {}.'.format(con, feature_name))
        else:
            raise RuntimeError('Required feature {} not found.'.format(feature_name))

    event = ProjectEvent(config=config)
    event.variables, event.files = variables, files

    # todo: add listener dump list in debug/verbose mode ?

    event = dispatcher.dispatch('medikit.on_start', event)

    dispatcher.dispatch('medikit.on_end', event)

    logger.info('Done.')


def _handle_pipeline_action(pipeline, action, *, filename, force=False):
    if action == START:
        if os.path.exists(filename):
            if force:
                os.unlink(filename)
            else:
                raise FileExistsError('Already started, use --force to force a restart, or use continue.')
        pipeline.init()
        with open(filename, 'w+') as f:
            f.write(pipeline.serialize())
        return CONTINUE
    elif action == CONTINUE:
        assert os.path.exists(filename)
        with open(filename) as f:
            pipeline.unserialize(f.read())

        try:
            step = pipeline.next()
            step.logger.info('Running {}.'.format(step))
            step.run(pipeline.meta)
            if step.complete:
                step.logger.info('{} is complete, moving forward.'.format(step))
            else:
                step.logger.warning('{} is NOT complete after run, exiting.'.format(step))
                return

        except StopIteration:
            return

        with open(filename, 'w+') as f:
            f.write(pipeline.serialize())

        return CONTINUE
    elif action == ABORT:
        assert os.path.exists(filename)
        try:
            with open(filename) as f:
                pipeline.unserialize(f.read())
            pipeline.abort()
        finally:
            os.unlink(filename)
    else:
        return


def handle_pipeline(config_filename, *, pipeline, action, force=False, verbose=False):
    dispatcher = LoggingDispatcher()
    variables, features, files, config = _read_configuration(dispatcher, config_filename)

    if not pipeline in config.pipelines:
        raise ValueError(
            'Undefined pipeline {!r}. Valid choices are: {}.'.format(
                pipeline, ', '.join(sorted(config.pipelines.keys()))
            )
        )
    pipeline = ConfiguredPipeline(pipeline, config.pipelines[pipeline])
    path = os.path.dirname(config_filename)
    pipeline_file = os.path.join(path, '.medikit-pipeline')

    while action:
        action = _handle_pipeline_action(pipeline, action, filename=pipeline_file, force=force)
        force = False



import textwrap


class Script(object):
    def __init__(self, script=None):
        self.script = self.parse_script(script)

    def parse_script(self, script):
        if not script:
            return []
        script = textwrap.dedent(str(script)).strip()
        return script.split('\n')

    def __iter__(self):
        for line in self.script:
            yield line

    def __str__(self):
        return '\n'.join(self.__iter__())


import json
import os

from git import Repo

from medikit.steps import Step


class Make(Step):
    def __init__(self, target):
        super().__init__()
        self.target = target
        self.__args__ = (target, )

    def run(self, meta):
        self.exec('make ' + self.target)
        self.set_complete()


class System(Step):
    def __init__(self, cmd):
        super().__init__()
        self.cmd = cmd
        self.__args__ = (cmd, )

    def run(self, meta):
        os.system(self.cmd)
        self.set_complete()


class Commit(Step):
    def __init__(self, message, *, tag=False):
        super().__init__()
        self.message = message
        self.tag = bool(tag)
        self.__args__ = (message, self.tag)

    def run(self, meta):
        branch = self.exec('git rev-parse --abbrev-ref HEAD')
        version = self.exec('python setup.py --version')
        assert version == meta['version']
        os.system('git commit -m ' + json.dumps(self.message.format(**meta)))
        if self.tag:
            os.system('git tag -am {version} {version}'.format(**meta))

        repo = Repo()
        for remote in repo.remotes:
            if str(remote) in ('origin', 'upstream'):
                self.logger.info('git push {} {}...'.format(remote, branch))
                os.system('git push {} {} --tags'.format(remote, branch))

        self.set_complete()



from medikit.steps import Step


def _pip_install(*reqs):
    """Install requirements given a path to requirem."""
    import importlib
    import pip

    pip.main(['install', '--upgrade', *reqs])

    # Some shenanigans to be sure everything is importable after this, especially .egg-link files which
    # are referenced in *.pth files and apparently loaded by site.py at some magic bootstrap moment of the
    # python interpreter.
    pip.utils.pkg_resources = importlib.reload(pip.utils.pkg_resources)
    import site
    importlib.reload(site)


class Install(Step):
    def run(self, meta):
        _pip_install('pip', 'wheel', 'twine')
        self.set_complete()



#!/usr/bin/env python
# coding: utf-8

import argparse
import logging
import os
import sys

import mondrian
from medikit.commands import START, CONTINUE, ABORT, handle_init, handle_update, handle_pipeline


def main(args=None):
    mondrian.setup(excepthook=True)
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    parser = argparse.ArgumentParser()

    parser.add_argument('--config', '-c', default='Projectfile')
    parser.add_argument('--verbose', '-v', action='store_true', default=False)

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    # todo aliases for update/init
    # warning: http://bugs.python.org/issue9234

    parser_init = subparsers.add_parser('init', help='Initialize a new project.')
    parser_init.set_defaults(handler=handle_init)
    parser_init.add_argument('target')
    parser_init.add_argument('--name')
    parser_init.add_argument('--description')
    parser_init.add_argument('--license')
    parser_init.add_argument('--feature', '-f', action='append', dest='features')

    parser_update = subparsers.add_parser('update', help='Update current project.')
    parser_update.set_defaults(handler=handle_update)

    parser_pipeline = subparsers.add_parser('pipeline', help='Execute multi-steps pipelines (release, etc.).')
    parser_pipeline.set_defaults(handler=handle_pipeline)
    parser_pipeline.add_argument('pipeline')
    parser_pipeline.add_argument(
        'action', choices=(
            START,
            CONTINUE,
            ABORT,
        )
    )
    parser_pipeline.add_argument('--force', '-f', action='store_true')

    options, more_args = parser.parse_known_args(args if args is not None else sys.argv[1:])
    if options.verbose:
        logger.setLevel(logging.DEBUG)

    options = vars(options)
    command, handler = options.pop('command'), options.pop('handler')

    config_filename = os.path.join(os.getcwd(), options.pop('target', '.'), options.pop('config'))

    if len(more_args):
        return handler(config_filename, more=more_args, **options)
    else:
        return handler(config_filename, **options)


if __name__ == '__main__':
    main()
